@@ -50,7 +50,7 @@ pub fn gen_com_interface(def: &TypeDef, gen: &Gen, include: TypeInclude) -> Toke
                             ::windows::Param::Owned(::std::convert::Into::<::windows::IUnknown>::into(self))
                         }
                     }
-                    impl<'a> ::windows::IntoParam<'a, ::windows::IUnknown> for &'a #name {
+                    impl<'a> ::windows::IntoParam<'a, ::windows::IUnknown> for &#name {
                         fn into_param(self) -> ::windows::Param<'a, ::windows::IUnknown> {
                             ::windows::Param::Owned(::std::convert::Into::<::windows::IUnknown>::into(::std::clone::Clone::clone(self)))
                         }
@@ -76,7 +76,7 @@ pub fn gen_com_interface(def: &TypeDef, gen: &Gen, include: TypeInclude) -> Toke
                                 ::windows::Param::Owned(::std::convert::Into::<#into>::into(self))
                             }
                         }
-                        impl<'a> ::windows::IntoParam<'a, #into> for &'a #name {
+                        impl<'a> ::windows::IntoParam<'a, #into> for &#name {
                             fn into_param(self) -> ::windows::Param<'a, #into> {
                                 ::windows::Param::Owned(::std::convert::Into::<#into>::into(::std::clone::Clone::clone(self)))
                             }
@@ -160,76 +160,65 @@ fn gen_method(
         to_ident(name)
     };
 
-    if signature.has_query_interface() {
-        let leading_params = &signature.params[..signature.params.len() - 2];
-        let params = gen_win32_params(leading_params, gen);
-        let args = leading_params.iter().map(|p| gen_win32_abi_arg(p));
+    match signature.kind() {
+        SignatureKind::QueryInterface => {
+            let leading_params = &signature.params[..signature.params.len() - 2];
+            let args = leading_params.iter().map(|p| gen_win32_abi_arg(p));
+            let params = gen_win32_params(leading_params, gen);
 
-        quote! {
-            pub unsafe fn #name<#constraints T: ::windows::Interface>(&self, #params) -> ::windows::Result<T> {
-                let mut result__ = ::std::option::Option::None;
-                (::windows::Interface::vtable(self).#vtable_offset)(::windows::Abi::abi(self), #(#args,)* &<T as ::windows::Interface>::IID, ::windows::Abi::set_abi(&mut result__)).and_some(result__)
+            quote! {
+                pub unsafe fn #name<#constraints T: ::windows::Interface>(&self, #params) -> ::windows::Result<T> {
+                    let mut result__ = ::std::option::Option::None;
+                    (::windows::Interface::vtable(self).#vtable_offset)(::std::mem::transmute_copy(self), #(#args,)* &<T as ::windows::Interface>::IID, &mut result__ as *mut _ as *mut _).and_some(result__)
+                }
             }
         }
-    } else if signature.has_retval() {
-        let leading_params = &signature.params[..signature.params.len() - 1];
-        let params = gen_win32_params(leading_params, gen);
-        let args = leading_params.iter().map(|p| gen_win32_abi_arg(p));
+        SignatureKind::ResultValue => {
+            let leading_params = &signature.params[..signature.params.len() - 1];
+            let args = leading_params.iter().map(|p| gen_win32_abi_arg(p));
+            let params = gen_win32_params(leading_params, gen);
+            let return_type_tokens = gen_win32_result_type(&signature, gen);
 
-        let mut return_param = signature.params[signature.params.len() - 1].clone();
-
-        let return_type_tokens = if return_param.signature.pointers > 1 {
-            return_param.signature.pointers -= 1;
-            gen_win32_param(&return_param, gen)
-        } else {
-            gen_name(&return_param.signature.kind, gen)
-        };
-
-        quote! {
-            pub unsafe fn #name<#constraints>(&self, #params) -> ::windows::Result<#return_type_tokens> {
-                let mut result__: <#return_type_tokens as ::windows::Abi>::Abi = ::std::mem::zeroed();
-                (::windows::Interface::vtable(self).#vtable_offset)(::windows::Abi::abi(self), #(#args,)* &mut result__)
-                .from_abi::<#return_type_tokens>(result__ )
+            quote! {
+                pub unsafe fn #name<#constraints>(&self, #params) -> ::windows::Result<#return_type_tokens> {
+                    let mut result__: <#return_type_tokens as ::windows::Abi>::Abi = ::std::mem::zeroed();
+                    (::windows::Interface::vtable(self).#vtable_offset)(::std::mem::transmute_copy(self), #(#args,)* &mut result__)
+                    .from_abi::<#return_type_tokens>(result__ )
+                }
             }
         }
-    } else if signature.has_udt_return() {
-        let params = gen_win32_params(&signature.params, gen);
-        let args = signature.params.iter().map(|p| gen_win32_abi_arg(p));
-        let return_type = gen_abi_type_name(&signature.return_type.unwrap().kind, gen);
+        SignatureKind::ResultVoid => {
+            let params = gen_win32_params(&signature.params, gen);
+            let args = signature.params.iter().map(|p| gen_win32_abi_arg(p));
 
-        quote! {
-            pub unsafe fn #name<#constraints>(&self, #params) -> #return_type {
-                let mut result__: #return_type = ::std::default::Default::default();
-                (::windows::Interface::vtable(self).#vtable_offset)(::windows::Abi::abi(self), #(#args,)* &mut result__);
-                result__
-            }
-        }
-    } else if let Some(return_type) = &signature.return_type {
-        let params = gen_win32_params(&signature.params, gen);
-        let args = signature.params.iter().map(|p| gen_win32_abi_arg(p));
-
-        if return_type.kind == ElementType::HRESULT {
             quote! {
                 pub unsafe fn #name<#constraints>(&self, #params) -> ::windows::Result<()> {
-                    (::windows::Interface::vtable(self).#vtable_offset)(::windows::Abi::abi(self), #(#args,)*).ok()
-                }
-            }
-        } else {
-            let return_type = gen_abi_sig(return_type, gen);
-
-            quote! {
-                pub unsafe fn #name<#constraints>(&self, #params) -> #return_type {
-                    (::windows::Interface::vtable(self).#vtable_offset)(::windows::Abi::abi(self), #(#args,)*)
+                    (::windows::Interface::vtable(self).#vtable_offset)(::std::mem::transmute_copy(self), #(#args,)*).ok()
                 }
             }
         }
-    } else {
-        let params = gen_win32_params(&signature.params, gen);
-        let args = signature.params.iter().map(|p| gen_win32_abi_arg(p));
+        SignatureKind::ReturnStruct => {
+            let params = gen_win32_params(&signature.params, gen);
+            let args = signature.params.iter().map(|p| gen_win32_abi_arg(p));
+            let return_sig = gen_abi_type_name(&signature.return_sig.unwrap().kind, gen);
 
-        quote! {
-            pub unsafe fn #name<#constraints>(&self, #params) {
-                (::windows::Interface::vtable(self).#vtable_offset)(::windows::Abi::abi(self), #(#args,)*)
+            quote! {
+                pub unsafe fn #name<#constraints>(&self, #params) -> #return_sig {
+                    let mut result__: #return_sig = ::std::default::Default::default();
+                    (::windows::Interface::vtable(self).#vtable_offset)(::std::mem::transmute_copy(self), #(#args,)* &mut result__);
+                    result__
+                }
+            }
+        }
+        SignatureKind::PreserveSig => {
+            let params = gen_win32_params(&signature.params, gen);
+            let args = signature.params.iter().map(|p| gen_win32_abi_arg(p));
+            let return_sig = gen_win32_return_sig(&signature, gen);
+
+            quote! {
+                pub unsafe fn #name<#constraints>(&self, #params) #return_sig {
+                    ::std::mem::transmute((::windows::Interface::vtable(self).#vtable_offset)(::std::mem::transmute_copy(self), #(#args,)*))
+                }
             }
         }
     }
