@@ -6,8 +6,8 @@ pub fn gen_winrt_upcall(sig: &MethodSignature, inner: TokenStream, gen: &Gen) ->
         .iter()
         .map(|param| gen_winrt_invoke_arg(param, gen));
 
-    match &sig.return_type {
-        Some(return_type) if return_type.is_array => {
+    match &sig.return_sig {
+        Some(return_sig) if return_sig.is_array => {
             quote! {
                 match #inner(#(#invoke_args,)*) {
                     ::std::result::Result::Ok(ok__) => {
@@ -57,8 +57,7 @@ pub fn gen_winrt_abi(sig: &MethodSignature, gen: &Gen) -> TokenStream {
                     quote! { #abi_size_name: u32, #name: *mut #abi }
                 }
             } else if p.param.is_input() {
-                // WinRT only uses const to mean that structs are passed by reference.
-                if p.is_const() {
+                if p.signature.is_const {
                     quote! { #name: &#abi }
                 } else {
                     quote! { #name: #abi }
@@ -67,7 +66,7 @@ pub fn gen_winrt_abi(sig: &MethodSignature, gen: &Gen) -> TokenStream {
                 quote! { #name: *mut #abi }
             }
         })
-        .chain(sig.return_type.iter().map(|signature| {
+        .chain(sig.return_sig.iter().map(|signature| {
             let abi = gen_abi_sig(signature, gen);
 
             if signature.is_array {
@@ -86,15 +85,20 @@ pub fn gen_winrt_invoke_arg(param: &MethodParam, gen: &Gen) -> TokenStream {
     let name = gen_param_name(&param.param);
     let kind = gen_name(&param.signature.kind, gen);
 
-    // TODO: This compiles but doesn't property handle delegates with array parameters.
-    // https://github.com/microsoft/windows-rs/issues/212
-
     if param.signature.is_array {
-        quote! { ::std::mem::transmute_copy(&#name) }
+        let abi_size_name = to_ident(&format!("{}_array_size", param.param.name()));
+
+        if param.param.is_input() {
+            quote! { ::std::slice::from_raw_parts(::std::mem::transmute_copy(&#name), #abi_size_name as _) }
+        } else if param.signature.by_ref {
+            quote! { ::windows::ArrayProxy::from_raw_parts(::std::mem::transmute_copy(&#name), #abi_size_name).as_array() }
+        } else {
+            quote! { ::std::slice::from_raw_parts_mut(::std::mem::transmute_copy(&#name), #abi_size_name as _) }
+        }
     } else if param.param.is_input() {
         if param.signature.kind.is_primitive() {
             quote! { #name }
-        } else if param.is_const() {
+        } else if param.signature.is_const {
             quote! { &*(#name as *const <#kind as ::windows::Abi>::Abi as *const <#kind as ::windows::Abi>::DefaultType) }
         } else {
             quote! { &*(&#name as *const <#kind as ::windows::Abi>::Abi as *const <#kind as ::windows::Abi>::DefaultType) }
@@ -124,18 +128,7 @@ pub fn gen_winrt_params(params: &[MethodParam], gen: &Gen) -> TokenStream {
                     let into = gen_name(&param.signature.kind, gen);
                     quote! { #name: impl ::windows::IntoParam<'a, #into>, }
                 } else {
-                    let mut signature = quote! {};
-
-                    for _ in 0..param.signature.pointers {
-                        if param.is_const() {
-                            signature.combine(&quote! { *const });
-                        } else {
-                            signature.combine(&quote! { *mut });
-                        }
-                    }
-
-                    signature.combine(&tokens);
-                    quote! { #name: #signature, }
+                    quote! { #name: #tokens, }
                 }
             } else if param.signature.kind.is_nullable() {
                 quote! { #name: &mut ::std::option::Option<#tokens>, }
@@ -181,10 +174,10 @@ pub fn gen_winrt_method(
     let params = gen_winrt_params(params, gen);
     let interface_name = gen_type_name(&interface.def, gen);
 
-    let return_type_tokens = if let Some(return_type) = &sig.return_type {
-        let tokens = gen_name(&return_type.kind, gen);
+    let return_type_tokens = if let Some(return_sig) = &sig.return_sig {
+        let tokens = gen_name(&return_sig.kind, gen);
 
-        if return_type.is_array {
+        if return_sig.is_array {
             quote! { ::windows::Array<#tokens> }
         } else {
             tokens
@@ -193,10 +186,10 @@ pub fn gen_winrt_method(
         quote! { () }
     };
 
-    let return_arg = if let Some(return_type) = &sig.return_type {
-        if return_type.is_array {
-            let return_type = gen_name(&return_type.kind, gen);
-            quote! { ::windows::Array::<#return_type>::set_abi_len(&mut result__), ::windows::Array::<#return_type>::set_abi(&mut result__) }
+    let return_arg = if let Some(return_sig) = &sig.return_sig {
+        if return_sig.is_array {
+            let return_sig = gen_name(&return_sig.kind, gen);
+            quote! { ::windows::Array::<#return_sig>::set_abi_len(&mut result__), &mut result__ as *mut _ as _ }
         } else {
             quote! { &mut result__ }
         }
@@ -208,31 +201,31 @@ pub fn gen_winrt_method(
     // arguments to ensure the call succeeds in the non-aggregating case.
     let composable_args = match interface.kind {
         InterfaceKind::Composable => quote! {
-            ::std::ptr::null_mut(), ::windows::Abi::set_abi(&mut ::std::option::Option::<::windows::IInspectable>::None),
+            ::std::ptr::null_mut(), &mut ::std::option::Option::<::windows::IInspectable>::None as *mut _ as _,
         },
         InterfaceKind::Extend => quote! {
-            ::windows::Abi::abi(&derived__), ::windows::Abi::set_abi(base__),
+            ::std::mem::transmute_copy(&derived__), base__ as *mut _ as _,
         },
         _ => quote! {},
     };
 
-    let vcall = if let Some(return_type) = &sig.return_type {
-        if return_type.is_array {
+    let vcall = if let Some(return_sig) = &sig.return_sig {
+        if return_sig.is_array {
             quote! {
                 let mut result__: #return_type_tokens = ::std::mem::zeroed();
-                (::windows::Interface::vtable(this).#vtable_offset)(::windows::Abi::abi(this), #(#args,)* #composable_args #return_arg)
+                (::windows::Interface::vtable(this).#vtable_offset)(::std::mem::transmute_copy(this), #(#args,)* #composable_args #return_arg)
                     .and_then(|| result__ )
             }
         } else {
             quote! {
                 let mut result__: <#return_type_tokens as ::windows::Abi>::Abi = ::std::mem::zeroed();
-                    (::windows::Interface::vtable(this).#vtable_offset)(::windows::Abi::abi(this), #(#args,)* #composable_args #return_arg)
+                    (::windows::Interface::vtable(this).#vtable_offset)(::std::mem::transmute_copy(this), #(#args,)* #composable_args #return_arg)
                         .from_abi::<#return_type_tokens>(result__ )
             }
         }
     } else {
         quote! {
-            (::windows::Interface::vtable(this).#vtable_offset)(::windows::Abi::abi(this), #(#args,)* #composable_args).ok()
+            (::windows::Interface::vtable(this).#vtable_offset)(::std::mem::transmute_copy(this), #(#args,)* #composable_args).ok()
         }
     };
 
@@ -256,7 +249,7 @@ pub fn gen_winrt_method(
             quote! {
                 #deprecated
                 pub fn #name<#constraints>(&self, #params) -> ::windows::Result<#return_type_tokens> {
-                    let this = &::windows::Interface::cast::<#interface_name>(self).unwrap();
+                    let this = &::windows::Interface::cast::<#interface_name>(self)?;
                     unsafe {
                         #vcall
                     }
@@ -292,13 +285,13 @@ pub fn gen_winrt_abi_arg(param: &MethodParam) -> TokenStream {
         if param.param.is_input() {
             quote! { #name.len() as u32, ::std::mem::transmute(#name.as_ptr()) }
         } else if param.signature.by_ref {
-            quote! { #name.set_abi_len(), #name.set_abi() }
+            quote! { #name.set_abi_len(), #name as *mut _ as _ }
         } else {
             quote! { #name.len() as u32, ::std::mem::transmute_copy(&#name) }
         }
     } else if param.param.is_input() {
         if param.is_convertible() {
-            if param.is_const() {
+            if param.signature.is_const {
                 quote! { &#name.into_param().abi() }
             } else {
                 quote! { #name.into_param().abi() }
@@ -306,7 +299,7 @@ pub fn gen_winrt_abi_arg(param: &MethodParam) -> TokenStream {
         } else if param.signature.kind.is_blittable() {
             quote! { #name }
         } else if param.signature.pointers == 0 {
-            quote! { ::windows::Abi::abi(#name) }
+            quote! { ::std::mem::transmute_copy(#name) }
         } else {
             quote! { ::std::mem::transmute(#name) }
         }
@@ -315,7 +308,7 @@ pub fn gen_winrt_abi_arg(param: &MethodParam) -> TokenStream {
     {
         quote! { #name }
     } else {
-        quote! { ::windows::Abi::set_abi(#name) }
+        quote! { #name as *mut _ as _ }
     }
 }
 
@@ -324,11 +317,11 @@ pub fn gen_winrt_produce_type(param: &MethodParam, gen: &Gen) -> TokenStream {
 
     if param.signature.is_array {
         if param.param.is_input() {
-            quote! { &[#tokens] }
+            quote! { &[<#tokens as ::windows::Abi>::DefaultType] }
         } else if param.signature.by_ref {
             quote! { &mut ::windows::Array<#tokens> }
         } else {
-            quote! { &mut [#tokens] }
+            quote! { &mut [<#tokens as ::windows::Abi>::DefaultType] }
         }
     } else if param.param.is_input() {
         if let ElementType::GenericParam(_) = param.signature.kind {
@@ -340,7 +333,26 @@ pub fn gen_winrt_produce_type(param: &MethodParam, gen: &Gen) -> TokenStream {
         } else {
             quote! { &#tokens }
         }
+    } else if param.signature.kind.is_nullable() {
+        quote! { &mut ::std::option::Option<#tokens> }
     } else {
         quote! { &mut #tokens }
     }
 }
+
+pub fn gen_phantoms(def: &TypeDef) -> impl Iterator<Item = TokenStream> + '_ {
+    def.generics.iter().map(move |g| {
+        let g = gen_name(g, &Gen::Absolute);
+        quote! { ::std::marker::PhantomData::<#g> }
+    })
+}
+
+pub fn gen_constraints(def: &TypeDef) -> TokenStream {
+    def.generics
+        .iter()
+        .map(|g| {
+            let g = gen_name(g, &Gen::Absolute);
+            quote! { #g: ::windows::RuntimeType + 'static, }
+        })
+        .collect()
+}
